from enum import Enum, auto
from typing import List, Union, Dict, Any
from math import inf
import sys

class Model(dict):
    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return str({key: (str(value)[:32] + '[...]') if len(str(value)) > 32 else value
                    for key, value in self.items()
                    if value is not None and not isinstance(value, (dict, list))})
        
class MetadataFieldDefinition(Model):
    def __init__(self, name: 'str' = None, description: 'str' = None, type: 'Dict' = None):
        super(MetadataFieldDefinition, self).__init__()
        self['name'] = name or None
        self['description'] = description or None
        self['type'] = type or None

    @property
    def name(self) -> 'str':
        return self['name']

    @name.setter
    def name(self, name: 'str'):
        self['name'] = name
            
    @property
    def description(self) -> 'str':
        return self['description']

    @description.setter
    def description(self, description: 'str'):
        self['description'] = description
            
    @property
    def type(self) -> 'Dict':
        return self['type']

    @type.setter
    def type(self, type: 'Dict'):
        self['type'] = type
            
    MEMBERS = {
        'name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'description': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'type': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class EntitiyDefinition(Model):
    def __init__(self, key: 'str' = None, name: 'str' = None, entity: 'str' = None, description: 'str' = None, type: 'Dict' = None):
        super(EntitiyDefinition, self).__init__()
        self['key'] = key or None
        self['name'] = name or None
        self['entity'] = entity or None
        self['description'] = description or None
        self['type'] = type or None

    @property
    def key(self) -> 'str':
        return self['key']

    @key.setter
    def key(self, key: 'str'):
        self['key'] = key
            
    @property
    def name(self) -> 'str':
        return self['name']

    @name.setter
    def name(self, name: 'str'):
        self['name'] = name
            
    @property
    def entity(self) -> 'str':
        return self['entity']

    @entity.setter
    def entity(self, entity: 'str'):
        self['entity'] = entity
            
    @property
    def description(self) -> 'str':
        return self['description']

    @description.setter
    def description(self, description: 'str'):
        self['description'] = description
            
    @property
    def type(self) -> 'Dict':
        return self['type']

    @type.setter
    def type(self, type: 'Dict'):
        self['type'] = type
            
    MEMBERS = {
        'key': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'entity': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'description': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'type': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class SuffixDefinition(Model):
    def __init__(self, name: 'str' = None, description: 'str' = None, type: 'Dict' = None):
        super(SuffixDefinition, self).__init__()
        self['name'] = name or None
        self['description'] = description or None
        self['type'] = type or None

    @property
    def name(self) -> 'str':
        return self['name']

    @name.setter
    def name(self, name: 'str'):
        self['name'] = name
            
    @property
    def description(self) -> 'str':
        return self['description']

    @description.setter
    def description(self, description: 'str'):
        self['description'] = description
            
    @property
    def type(self) -> 'Dict':
        return self['type']

    @type.setter
    def type(self, type: 'Dict'):
        self['type'] = type
            
    MEMBERS = {
        'name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'description': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'type': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class File(Model):
    def __init__(self, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(File, self).__init__()
        self['name'] = name or None
        self['extension'] = extension or None
        self['uri'] = uri or None

    @property
    def name(self) -> 'str':
        return self['name']

    @name.setter
    def name(self, name: 'str'):
        self['name'] = name
            
    @property
    def extension(self) -> 'str':
        return self['extension']

    @extension.setter
    def extension(self, extension: 'str'):
        self['extension'] = extension
            
    @property
    def uri(self) -> 'str':
        return self['uri']

    @uri.setter
    def uri(self, uri: 'str'):
        self['uri'] = uri
            
    MEMBERS = {
        'name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'extension': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'uri': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class JsonFile(File):
    def __init__(self, contents: 'Dict' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(JsonFile, self).__init__(name or None, extension or None, uri or None)
        self['contents'] = contents or None

    @property
    def contents(self) -> 'Dict':
        return self['contents']

    @contents.setter
    def contents(self, contents: 'Dict'):
        self['contents'] = contents
            
    MEMBERS = {
        'contents': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class Artifact(File):
    r"""An artifact is a file whose name conforms to the BIDS file naming convention."""
    def __init__(self, suffix: 'str' = None, datatype: 'str' = None, entities: 'List[EntityRef]' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(Artifact, self).__init__(name or None, extension or None, uri or None)
        self['suffix'] = suffix or None
        self['datatype'] = datatype or None
        self['entities'] = entities or []

    @property
    def suffix(self) -> 'str':
        return self['suffix']

    @suffix.setter
    def suffix(self, suffix: 'str'):
        self['suffix'] = suffix
            
    @property
    def datatype(self) -> 'str':
        return self['datatype']

    @datatype.setter
    def datatype(self, datatype: 'str'):
        self['datatype'] = datatype
            
    @property
    def entities(self) -> 'List[EntityRef]':
        return self['entities']

    @entities.setter
    def entities(self, entities: 'List[EntityRef]'):
        self['entities'] = entities
            
    MEMBERS = {
        'suffix': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'datatype': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'entities': {'type': 'EntityRef', 'min': 1, 'max': inf, 'use': 'optional', 'meta': {}},
    }


class MetadataArtifact(Artifact):
    def __init__(self, contents: 'Dict' = None, suffix: 'str' = None, datatype: 'str' = None, entities: 'List[EntityRef]' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(MetadataArtifact, self).__init__(suffix or None, datatype or None, entities or [], name or None, extension or None, uri or None)
        self['contents'] = contents or None

    @property
    def contents(self) -> 'Dict':
        return self['contents']

    @contents.setter
    def contents(self, contents: 'Dict'):
        self['contents'] = contents
            
    MEMBERS = {
        'contents': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class MetadataFile(File):
    def __init__(self, contents: 'Dict' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(MetadataFile, self).__init__(name or None, extension or None, uri or None)
        self['contents'] = contents or None

    @property
    def contents(self) -> 'Dict':
        return self['contents']

    @contents.setter
    def contents(self, contents: 'Dict'):
        self['contents'] = contents
            
    MEMBERS = {
        'contents': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class TSVArtifact(Artifact):
    def __init__(self, delimiter: 'str' = None, contents: 'Dict' = None, suffix: 'str' = None, datatype: 'str' = None, entities: 'List[EntityRef]' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(TSVArtifact, self).__init__(suffix or None, datatype or None, entities or [], name or None, extension or None, uri or None)
        self['delimiter'] = delimiter or None
        self['contents'] = contents or None

    @property
    def delimiter(self) -> 'str':
        return self['delimiter']

    @delimiter.setter
    def delimiter(self, delimiter: 'str'):
        self['delimiter'] = delimiter
            
    @property
    def contents(self) -> 'Dict':
        return self['contents']

    @contents.setter
    def contents(self, contents: 'Dict'):
        self['contents'] = contents
            
    MEMBERS = {
        'delimiter': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'contents': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class TSVFile(File):
    def __init__(self, delimiter: 'str' = None, contents: 'Dict' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(TSVFile, self).__init__(name or None, extension or None, uri or None)
        self['delimiter'] = delimiter or None
        self['contents'] = contents or None

    @property
    def delimiter(self) -> 'str':
        return self['delimiter']

    @delimiter.setter
    def delimiter(self, delimiter: 'str'):
        self['delimiter'] = delimiter
            
    @property
    def contents(self) -> 'Dict':
        return self['contents']

    @contents.setter
    def contents(self, contents: 'Dict'):
        self['contents'] = contents
            
    MEMBERS = {
        'delimiter': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'contents': {'type': 'dict', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class Folder(Model):
    def __init__(self, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(Folder, self).__init__()
        self['name'] = name or None
        self['files'] = files or []
        self['folders'] = folders or []

    @property
    def name(self) -> 'str':
        return self['name']

    @name.setter
    def name(self, name: 'str'):
        self['name'] = name
            
    @property
    def files(self) -> 'List[File]':
        return self['files']

    @files.setter
    def files(self, files: 'List[File]'):
        self['files'] = files
            
    @property
    def folders(self) -> 'List[Folder]':
        return self['folders']

    @folders.setter
    def folders(self, folders: 'List[Folder]'):
        self['folders'] = folders
            
    MEMBERS = {
        'name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'files': {'type': 'File', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
        'folders': {'type': 'Folder', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
    }


class EntityRef(Model):
    def __init__(self, key: 'str' = None, value: 'str' = None):
        super(EntityRef, self).__init__()
        self['key'] = key or None
        self['value'] = value or None

    @property
    def key(self) -> 'str':
        return self['key']

    @key.setter
    def key(self, key: 'str'):
        self['key'] = key
            
    @property
    def value(self) -> 'str':
        return self['value']

    @value.setter
    def value(self, value: 'str'):
        self['value'] = value
            
    MEMBERS = {
        'key': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'value': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
    }


class DatasetDescriptionFile(JsonFile):
    class DatasetTypeEnum(Enum):
        raw = auto()
        derivative = auto()

    def __init__(self, Name: 'str' = None, BIDSVersion: 'str' = None, HEDVersion: 'str' = None, DatasetType: 'DatasetDescriptionFile.DatasetTypeEnum' = None, License: 'str' = None, Acknowledgements: 'str' = None, HowToAcknowledge: 'str' = None, DatasetDOI: 'str' = None, Authors: 'List[str]' = None, Funding: 'List[str]' = None, EthicsApprovals: 'List[str]' = None, ReferencesAndLinks: 'List[str]' = None, contents: 'Dict' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(DatasetDescriptionFile, self).__init__(contents or None, name or None, extension or None, uri or None)
        self['Name'] = Name or None
        self['BIDSVersion'] = BIDSVersion or None
        self['HEDVersion'] = HEDVersion or None
        self['DatasetType'] = DatasetType or None
        self['License'] = License or None
        self['Acknowledgements'] = Acknowledgements or None
        self['HowToAcknowledge'] = HowToAcknowledge or None
        self['DatasetDOI'] = DatasetDOI or None
        self['Authors'] = Authors or []
        self['Funding'] = Funding or []
        self['EthicsApprovals'] = EthicsApprovals or []
        self['ReferencesAndLinks'] = ReferencesAndLinks or []

    @property
    def Name(self) -> 'str':
        return self['Name']

    @Name.setter
    def Name(self, Name: 'str'):
        self['Name'] = Name
            
    @property
    def BIDSVersion(self) -> 'str':
        return self['BIDSVersion']

    @BIDSVersion.setter
    def BIDSVersion(self, BIDSVersion: 'str'):
        self['BIDSVersion'] = BIDSVersion
            
    @property
    def HEDVersion(self) -> 'str':
        return self['HEDVersion']

    @HEDVersion.setter
    def HEDVersion(self, HEDVersion: 'str'):
        self['HEDVersion'] = HEDVersion
            
    @property
    def DatasetType(self) -> 'DatasetDescriptionFile.DatasetTypeEnum':
        r"""The interpretation of the dataset. MUST be one of "raw" or "derivative". For backwards compatibility, the default value is "raw"."""
        return self['DatasetType']

    @DatasetType.setter
    def DatasetType(self, DatasetType: 'DatasetDescriptionFile.DatasetTypeEnum'):
        self['DatasetType'] = DatasetType
            
    @property
    def License(self) -> 'str':
        return self['License']

    @License.setter
    def License(self, License: 'str'):
        self['License'] = License
            
    @property
    def Acknowledgements(self) -> 'str':
        return self['Acknowledgements']

    @Acknowledgements.setter
    def Acknowledgements(self, Acknowledgements: 'str'):
        self['Acknowledgements'] = Acknowledgements
            
    @property
    def HowToAcknowledge(self) -> 'str':
        return self['HowToAcknowledge']

    @HowToAcknowledge.setter
    def HowToAcknowledge(self, HowToAcknowledge: 'str'):
        self['HowToAcknowledge'] = HowToAcknowledge
            
    @property
    def DatasetDOI(self) -> 'str':
        return self['DatasetDOI']

    @DatasetDOI.setter
    def DatasetDOI(self, DatasetDOI: 'str'):
        self['DatasetDOI'] = DatasetDOI
            
    @property
    def Authors(self) -> 'List[str]':
        return self['Authors']

    @Authors.setter
    def Authors(self, Authors: 'List[str]'):
        self['Authors'] = Authors
            
    @property
    def Funding(self) -> 'List[str]':
        return self['Funding']

    @Funding.setter
    def Funding(self, Funding: 'List[str]'):
        self['Funding'] = Funding
            
    @property
    def EthicsApprovals(self) -> 'List[str]':
        return self['EthicsApprovals']

    @EthicsApprovals.setter
    def EthicsApprovals(self, EthicsApprovals: 'List[str]'):
        self['EthicsApprovals'] = EthicsApprovals
            
    @property
    def ReferencesAndLinks(self) -> 'List[str]':
        return self['ReferencesAndLinks']

    @ReferencesAndLinks.setter
    def ReferencesAndLinks(self, ReferencesAndLinks: 'List[str]'):
        self['ReferencesAndLinks'] = ReferencesAndLinks
            
    MEMBERS = {
        'Name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'BIDSVersion': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'HEDVersion': {'type': 'str', 'min': 0, 'max': 1, 'use': 'recommended', 'meta': {}},
        'DatasetType': {'type': 'DatasetTypeEnum', 'min': 0, 'max': 1, 'use': 'recommended', 'meta': {}},
        'License': {'type': 'str', 'min': 0, 'max': 1, 'use': 'recommended', 'meta': {}},
        'Acknowledgements': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'HowToAcknowledge': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'DatasetDOI': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'Authors': {'type': 'str', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
        'Funding': {'type': 'str', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
        'EthicsApprovals': {'type': 'str', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
        'ReferencesAndLinks': {'type': 'str', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
    }


class DerivativeDatasetDescriptionFile(DatasetDescriptionFile):
    def __init__(self, GeneratedBy: 'List[GeneratedBy]' = None, SourceDatasets: 'List[SourceDatasets]' = None, Name: 'str' = None, BIDSVersion: 'str' = None, HEDVersion: 'str' = None, DatasetType: 'DatasetDescriptionFile.DatasetTypeEnum' = None, License: 'str' = None, Acknowledgements: 'str' = None, HowToAcknowledge: 'str' = None, DatasetDOI: 'str' = None, Authors: 'List[str]' = None, Funding: 'List[str]' = None, EthicsApprovals: 'List[str]' = None, ReferencesAndLinks: 'List[str]' = None, contents: 'Dict' = None, name: 'str' = None, extension: 'str' = None, uri: 'str' = None):
        super(DerivativeDatasetDescriptionFile, self).__init__(Name or None, BIDSVersion or None, HEDVersion or None, DatasetType or None, License or None, Acknowledgements or None, HowToAcknowledge or None, DatasetDOI or None, Authors or [], Funding or [], EthicsApprovals or [], ReferencesAndLinks or [], contents or None, name or None, extension or None, uri or None)
        self['GeneratedBy'] = GeneratedBy or []
        self['SourceDatasets'] = SourceDatasets or []

    @property
    def GeneratedBy(self) -> 'List[GeneratedBy]':
        return self['GeneratedBy']

    @GeneratedBy.setter
    def GeneratedBy(self, GeneratedBy: 'List[GeneratedBy]'):
        self['GeneratedBy'] = GeneratedBy
            
    @property
    def SourceDatasets(self) -> 'List[SourceDatasets]':
        return self['SourceDatasets']

    @SourceDatasets.setter
    def SourceDatasets(self, SourceDatasets: 'List[SourceDatasets]'):
        self['SourceDatasets'] = SourceDatasets
            
    MEMBERS = {
        'GeneratedBy': {'type': 'GeneratedBy', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
        'SourceDatasets': {'type': 'SourceDatasets', 'min': 0, 'max': inf, 'use': 'recommended', 'meta': {}},
    }


class DerivativeFolder(Folder):
    def __init__(self, dataset_description: 'DerivativeDatasetDescriptionFile' = None, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(DerivativeFolder, self).__init__(name or None, files or [], folders or [])
        self['dataset_description'] = dataset_description or None

    @property
    def dataset_description(self) -> 'DerivativeDatasetDescriptionFile':
        return self['dataset_description']

    @dataset_description.setter
    def dataset_description(self, dataset_description: 'DerivativeDatasetDescriptionFile'):
        self['dataset_description'] = dataset_description
            
    MEMBERS = {
        'dataset_description': {'type': 'DerivativeDatasetDescriptionFile', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class SessionFolder(Folder):
    def __init__(self, datatypes: 'List[DatatypeFolder]' = None, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(SessionFolder, self).__init__(name or None, files or [], folders or [])
        self['datatypes'] = datatypes or []

    @property
    def datatypes(self) -> 'List[DatatypeFolder]':
        return self['datatypes']

    @datatypes.setter
    def datatypes(self, datatypes: 'List[DatatypeFolder]'):
        self['datatypes'] = datatypes
            
    MEMBERS = {
        'datatypes': {'type': 'DatatypeFolder', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
    }


class DatatypeFolder(Folder):
    def __init__(self, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(DatatypeFolder, self).__init__(name or None, files or [], folders or [])

    MEMBERS = {
    }


class Subject(Folder):
    def __init__(self, sessions: 'List[SessionFolder]' = None, datatypes: 'List[DatatypeFolder]' = None, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(Subject, self).__init__(name or None, files or [], folders or [])
        self['sessions'] = sessions or []
        self['datatypes'] = datatypes or []

    @property
    def sessions(self) -> 'List[SessionFolder]':
        return self['sessions']

    @sessions.setter
    def sessions(self, sessions: 'List[SessionFolder]'):
        self['sessions'] = sessions
            
    @property
    def datatypes(self) -> 'List[DatatypeFolder]':
        return self['datatypes']

    @datatypes.setter
    def datatypes(self, datatypes: 'List[DatatypeFolder]'):
        self['datatypes'] = datatypes
            
    MEMBERS = {
        'sessions': {'type': 'SessionFolder', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {'name_pattern': 'ses-.*'}},
        'datatypes': {'type': 'DatatypeFolder', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {'name_pattern': '.*'}},
    }


class GeneratedBy(Model):
    def __init__(self, Name: 'str' = None, Version: 'str' = None, Description: 'str' = None, CodeURL: 'str' = None, Container: 'List[GeneratedByContainer]' = None):
        super(GeneratedBy, self).__init__()
        self['Name'] = Name or None
        self['Version'] = Version or None
        self['Description'] = Description or None
        self['CodeURL'] = CodeURL or None
        self['Container'] = Container or []

    @property
    def Name(self) -> 'str':
        return self['Name']

    @Name.setter
    def Name(self, Name: 'str'):
        self['Name'] = Name
            
    @property
    def Version(self) -> 'str':
        return self['Version']

    @Version.setter
    def Version(self, Version: 'str'):
        self['Version'] = Version
            
    @property
    def Description(self) -> 'str':
        return self['Description']

    @Description.setter
    def Description(self, Description: 'str'):
        self['Description'] = Description
            
    @property
    def CodeURL(self) -> 'str':
        return self['CodeURL']

    @CodeURL.setter
    def CodeURL(self, CodeURL: 'str'):
        self['CodeURL'] = CodeURL
            
    @property
    def Container(self) -> 'List[GeneratedByContainer]':
        return self['Container']

    @Container.setter
    def Container(self, Container: 'List[GeneratedByContainer]'):
        self['Container'] = Container
            
    MEMBERS = {
        'Name': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'Version': {'type': 'str', 'min': 0, 'max': 1, 'use': 'recommended', 'meta': {}},
        'Description': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'CodeURL': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'Container': {'type': 'GeneratedByContainer', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {}},
    }


class SourceDatasets(Model):
    def __init__(self, DOI: 'str' = None, URL: 'str' = None, Version: 'str' = None):
        super(SourceDatasets, self).__init__()
        self['DOI'] = DOI or None
        self['URL'] = URL or None
        self['Version'] = Version or None

    @property
    def DOI(self) -> 'str':
        return self['DOI']

    @DOI.setter
    def DOI(self, DOI: 'str'):
        self['DOI'] = DOI
            
    @property
    def URL(self) -> 'str':
        return self['URL']

    @URL.setter
    def URL(self, URL: 'str'):
        self['URL'] = URL
            
    @property
    def Version(self) -> 'str':
        return self['Version']

    @Version.setter
    def Version(self, Version: 'str'):
        self['Version'] = Version
            
    MEMBERS = {
        'DOI': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'URL': {'type': 'str', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'Version': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class GeneratedByContainer(Model):
    def __init__(self, Type: 'str' = None, Tag: 'str' = None, URI: 'str' = None):
        super(GeneratedByContainer, self).__init__()
        self['Type'] = Type or None
        self['Tag'] = Tag or None
        self['URI'] = URI or None

    @property
    def Type(self) -> 'str':
        return self['Type']

    @Type.setter
    def Type(self, Type: 'str'):
        self['Type'] = Type
            
    @property
    def Tag(self) -> 'str':
        return self['Tag']

    @Tag.setter
    def Tag(self, Tag: 'str'):
        self['Tag'] = Tag
            
    @property
    def URI(self) -> 'str':
        return self['URI']

    @URI.setter
    def URI(self, URI: 'str'):
        self['URI'] = URI
            
    MEMBERS = {
        'Type': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'Tag': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'URI': {'type': 'str', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class Dataset(Folder):
    r"""The entry point of an in-memory graph representation of a BIDS dataset."""
    def __init__(self, subjects: 'List[Subject]' = None, dataset_description: 'DatasetDescriptionFile' = None, README: 'File' = None, CHANGES: 'File' = None, LICENSE: 'File' = None, genetic_info: 'JsonFile' = None, samples: 'JsonFile' = None, participants_tsv: 'File' = None, participants_json: 'JsonFile' = None, code: 'Folder' = None, derivatives: 'Folder' = None, sourcedata: 'Folder' = None, stimuli: 'Folder' = None, name: 'str' = None, files: 'List[File]' = None, folders: 'List[Folder]' = None):
        super(Dataset, self).__init__(name or None, files or [], folders or [])
        self['subjects'] = subjects or []
        self['dataset_description'] = dataset_description or None
        self['README'] = README or None
        self['CHANGES'] = CHANGES or None
        self['LICENSE'] = LICENSE or None
        self['genetic_info'] = genetic_info or None
        self['samples'] = samples or None
        self['participants_tsv'] = participants_tsv or None
        self['participants_json'] = participants_json or None
        self['code'] = code or None
        self['derivatives'] = derivatives or None
        self['sourcedata'] = sourcedata or None
        self['stimuli'] = stimuli or None

    @property
    def subjects(self) -> 'List[Subject]':
        return self['subjects']

    @subjects.setter
    def subjects(self, subjects: 'List[Subject]'):
        self['subjects'] = subjects
            
    @property
    def dataset_description(self) -> 'DatasetDescriptionFile':
        return self['dataset_description']

    @dataset_description.setter
    def dataset_description(self, dataset_description: 'DatasetDescriptionFile'):
        self['dataset_description'] = dataset_description
            
    @property
    def README(self) -> 'File':
        return self['README']

    @README.setter
    def README(self, README: 'File'):
        self['README'] = README
            
    @property
    def CHANGES(self) -> 'File':
        return self['CHANGES']

    @CHANGES.setter
    def CHANGES(self, CHANGES: 'File'):
        self['CHANGES'] = CHANGES
            
    @property
    def LICENSE(self) -> 'File':
        return self['LICENSE']

    @LICENSE.setter
    def LICENSE(self, LICENSE: 'File'):
        self['LICENSE'] = LICENSE
            
    @property
    def genetic_info(self) -> 'JsonFile':
        return self['genetic_info']

    @genetic_info.setter
    def genetic_info(self, genetic_info: 'JsonFile'):
        self['genetic_info'] = genetic_info
            
    @property
    def samples(self) -> 'JsonFile':
        return self['samples']

    @samples.setter
    def samples(self, samples: 'JsonFile'):
        self['samples'] = samples
            
    @property
    def participants_tsv(self) -> 'File':
        return self['participants_tsv']

    @participants_tsv.setter
    def participants_tsv(self, participants_tsv: 'File'):
        self['participants_tsv'] = participants_tsv
            
    @property
    def participants_json(self) -> 'JsonFile':
        return self['participants_json']

    @participants_json.setter
    def participants_json(self, participants_json: 'JsonFile'):
        self['participants_json'] = participants_json
            
    @property
    def code(self) -> 'Folder':
        return self['code']

    @code.setter
    def code(self, code: 'Folder'):
        self['code'] = code
            
    @property
    def derivatives(self) -> 'Folder':
        return self['derivatives']

    @derivatives.setter
    def derivatives(self, derivatives: 'Folder'):
        self['derivatives'] = derivatives
            
    @property
    def sourcedata(self) -> 'Folder':
        return self['sourcedata']

    @sourcedata.setter
    def sourcedata(self, sourcedata: 'Folder'):
        self['sourcedata'] = sourcedata
            
    @property
    def stimuli(self) -> 'Folder':
        return self['stimuli']

    @stimuli.setter
    def stimuli(self, stimuli: 'Folder'):
        self['stimuli'] = stimuli
            
    MEMBERS = {
        'subjects': {'type': 'Subject', 'min': 0, 'max': inf, 'use': 'optional', 'meta': {'name_pattern': 'sub-.*'}},
        'dataset_description': {'type': 'DatasetDescriptionFile', 'min': 0, 'max': 1, 'use': 'required', 'meta': {}},
        'README': {'type': 'File', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'CHANGES': {'type': 'File', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'LICENSE': {'type': 'File', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'genetic_info': {'type': 'JsonFile', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'samples': {'type': 'JsonFile', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'participants_tsv': {'type': 'File', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {'name_pattern': 'participants.tsv'}},
        'participants_json': {'type': 'JsonFile', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {'name_pattern': 'participants.json'}},
        'code': {'type': 'Folder', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'derivatives': {'type': 'Folder', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'sourcedata': {'type': 'Folder', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
        'stimuli': {'type': 'Folder', 'min': 0, 'max': 1, 'use': 'optional', 'meta': {}},
    }


class DatatypeEnum(Enum):
 pass
class ModalityEnum(Enum):
 pass
class SuffixEnum(Enum):
 pass
class EntityEnum(Enum):
 pass



from typing import List

from ancpbids.plugin import SchemaPlugin
from ancpbids.utils import deepupdate


def _get_metadata(artifact, include_entities=False):
    schema = artifact.get_schema()
    parent = artifact.get_parent()
    artifact_entities = {e.key : e.value for e in artifact.entities}
    # first, collect all metadata files matching the artifact
    metadata_levels = []
    while parent is not None:
        parent_metadata_files = parent.select(schema.MetadataArtifact).objects(depth=1)
        # check if a metadata file matches the artifact and merge its fields into the resulting metadata dict
        for parent_mdf in parent_metadata_files:
            match = parent_mdf.suffix == artifact.suffix
            if not match:
                continue
            mdf_entities = {e.key : e.value for e in parent_mdf.entities}
            match = mdf_entities.items() <= artifact_entities.items()
            if match:
                metadata_levels.append(parent_mdf.contents)
        parent = parent.get_parent()

    metadata = {}
    # now apply all metadata fields in reverse order, i.e. by starting from root down to the artifact level
    # this is important as the nearer the file is to the Artifact the higher its priority is
    for mdf in reversed(metadata_levels):
        deepupdate(metadata, mdf)

    if include_entities:
        schema_entities = {e.value['name']: e.name for e in list(schema.EntityEnum)}
        metadata.update({schema_entities[key]: value for key, value in artifact_entities.items()})

    return metadata


class MetadataSchemaPlugin(SchemaPlugin):
    def execute(self, schema):
        schema.Artifact.get_metadata = _get_metadata


import os.path

from ancpbids import load_dataset, DatasetOptions
from ancpbids.utils import parse_bids_name
from ..base_test_case import *


class BasicTestCase(BaseTestCase):
    def test_naming_scheme(self):
        valid_names = ["sub-11_task-mixedgamblestask_run-02_events.tsv", "sub-11_dwi.nii.gz", "x-01_y-02_z-03_xyz.abc",
                       "sub-01_task-mixedgamblestask_run-01_bold.nii.gz"]
        for name in valid_names:
            self.assertTrue(parse_bids_name(name) is not None)

        invalid_names = ["sub-04_T1w_bias.nii.gz", "cat.jpg", "readme.txt"]
        for name in invalid_names:
            self.assertTrue(parse_bids_name(name) is None)

    def test_parse_bids_name(self):
        bids_obj = parse_bids_name("sub-11_task-mixedgamblestask_run-02_bold.nii.gz")
        self.assertTrue(isinstance(bids_obj, dict))
        self.assertEqual(['sub', 'task', 'run'], list(bids_obj['entities'].keys()))
        self.assertEqual(['11', 'mixedgamblestask', '02'], list(bids_obj['entities'].values()))
        self.assertEqual('bold', bids_obj['suffix'])
        self.assertEqual('.nii.gz', bids_obj['extension'])

    def test_ds005_basic_structure(self):
        ds005 = load_dataset(DS005_DIR, DatasetOptions(lazy_loading=True))
        self.assertEqual("ds005", ds005.name)

        ds_descr = ds005.dataset_description
        self.assertTrue(isinstance(ds_descr, ds005.get_schema().DatasetDescriptionFile))
        self.assertEqual('dataset_description.json', ds_descr.name)
        self.assertEqual("1.0.0rc2", ds_descr.BIDSVersion)
        self.assertEqual("Mixed-gambles task", ds_descr.Name)
        self.assertTrue(ds_descr.License.startswith(
            "This dataset is made available under the Public Domain Dedication and License"))
        self.assertEqual([
            "Tom, S.M., Fox, C.R., Trepel, C., "
            "Poldrack, R.A. (2007). "
            "The neural basis of loss aversion in decision-making under risk. "
            "Science, 315(5811):515-8"],
            ds_descr.ReferencesAndLinks)

        subjects = ds005.subjects
        self.assertEqual(16, len(subjects))

        first_subject = subjects[0]
        self.assertEqual("sub-01", first_subject.name)
        last_subject = subjects[-1]
        self.assertEqual("sub-16", last_subject.name)

        sessions = first_subject.sessions
        self.assertEqual(0, len(sessions))

        datatypes = first_subject.datatypes
        self.assertEqual(3, len(datatypes))

        anat_datatype = datatypes[0]
        self.assertEqual("anat", anat_datatype.name)
        func_datatype = datatypes[-1]
        self.assertEqual("func", func_datatype.name)

        artifacts = func_datatype.query(suffix="bold", scope="self")
        self.assertEqual(3, len(artifacts))
        self.assertEqual("sub-01_task-mixedgamblestask_run-01_bold.nii.gz", artifacts[0].name)
        self.assertEqual("sub-01_task-mixedgamblestask_run-02_bold.nii.gz", artifacts[1].name)
        self.assertEqual("sub-01_task-mixedgamblestask_run-03_bold.nii.gz", artifacts[2].name)

        eventmetafiles = func_datatype.query(suffix="events", extension=".json", scope="self")
        self.assertEqual(1, len(eventmetafiles))
        self.assertEqual("sub-01_task-mixedgamblestask_run-01_events.json", eventmetafiles[0].name)

        tsvfiles = func_datatype.query(extension=".tsv", scope="self")
        self.assertEqual(3, len(tsvfiles))
        self.assertEqual("sub-01_task-mixedgamblestask_run-01_events.tsv", tsvfiles[0].name)
        self.assertEqual("sub-01_task-mixedgamblestask_run-02_events.tsv", tsvfiles[1].name)
        self.assertEqual("sub-01_task-mixedgamblestask_run-03_events.tsv", tsvfiles[2].name)

    def test_json_file_contents(self):
        ds005 = load_dataset(DS005_DIR)
        dataset_description = ds005.load_file_contents("dataset_description.json")
        self.assertTrue(isinstance(dataset_description, dict), "Expected a dictionary")
        self.assertEqual("1.0.0rc2", dataset_description['BIDSVersion'])
        self.assertEqual("Mixed-gambles task", dataset_description['Name'])

    def test_tsv_file_contents(self):
        ds005 = load_dataset(DS005_DIR)
        participants = ds005.load_file_contents("participants.tsv")
        self.assertListEqual(['participant_id', 'sex', 'age'], list(participants[0].keys()))
        self.assertEqual(16, len(participants))
        participants = ds005.load_file_contents("participants.tsv", return_type="ndarray")
        self.assertListEqual(['participant_id', 'sex', 'age'], list(participants.dtype.names))
        self.assertEqual(16, len(participants))
        participants = ds005.load_file_contents("participants.tsv", return_type="dataframe")
        self.assertListEqual(['participant_id', 'sex', 'age'], list(participants.columns))
        self.assertEqual(16, len(participants))

    def test_parse_entities_in_filenames(self):
        ds005 = load_dataset(DS005_DIR)
        # get first artifact in func datatype of first subject/session:
        # sub-16_task-mixedgamblesatask_run-01_bold.nii.gz
        artifact = ds005.subjects[0].datatypes[-1].query(scope="self")[0]
        self.assertTrue(isinstance(artifact, ds005.get_schema().Artifact))

        self.assertEqual("bold", artifact.suffix)
        self.assertEqual(".nii.gz", artifact.extension)

        entities = artifact.entities
        self.assertTrue(isinstance(entities, list))
        self.assertEqual(3, len(entities))

        entity = entities[0]
        self.assertEqual("sub", entity.key)
        self.assertEqual("01", entity.value)

        entity = entities[1]
        self.assertEqual("task", entity.key)
        self.assertEqual("mixedgamblestask", entity.value)

        entity = entities[2]
        self.assertEqual("run", entity.key)
        self.assertEqual(1, entity.value)

    def test_to_generator(self):
        ds005 = load_dataset(DS005_DIR)
        schema = ds005.get_schema()

        all_direct_files = list(
            ds005.to_generator(depth_first=True, depth=1, filter_=lambda n: isinstance(n, schema.File)))
        self.assertEqual(8, len(all_direct_files))

    def test_get_files_and_folders(self):
        ds005 = load_dataset(DS005_DIR)

        file = ds005.get_file("dataset_description.json")
        self.assertIsNotNone(file)
        self.assertEqual("dataset_description.json", file.name)

        file = ds005.get_file("sub-01/func/sub-01_task-mixedgamblestask_run-01_bold.nii.gz")
        self.assertIsNotNone(file)
        self.assertEqual("sub-01_task-mixedgamblestask_run-01_bold.nii.gz", file.name)

    def test_repr(self):
        ds005 = load_dataset(DS005_DIR)
        self.assertEqual("{'name': 'ds005'}", str(ds005))
        self.assertEqual("{'name': 'derivatives'}", str(ds005.derivatives))
        self.assertEqual("{'name': 'README'}", str(ds005.README))
        expected = "{'name': 'dataset_description.json', 'Name': 'Mixed-gambles task', 'BIDSVersion': '1.0.0rc2', 'License': 'This dataset is made available u[...]'}"
        self.assertEqual(expected, str(ds005.dataset_description))

    def test_participants_tsv(self):
        ds005 = load_dataset(DS005_DIR)
        schema = ds005.get_schema()
        self.assertTrue(isinstance(ds005.participants_tsv, schema.TSVFile))
        contents = ds005.participants_tsv.contents
        self.assertTrue(contents is not None)
        self.assertEqual(16, len(contents))
        self.assertEqual(['participant_id', 'sex', 'age'], list(contents[0].keys()))
        self.assertEqual(['sub-01', '0', '28'], list(contents[0].values()))

        # or short form:
        self.assertEqual({'participant_id': 'sub-01', 'sex': '0', 'age': '28'}, contents[0])

    def test_absolute_path(self):
        ds_path_norm = os.path.normpath(DS005_DIR)
        ds005 = load_dataset(ds_path_norm)
        ds_path = ds005.get_absolute_path()
        self.assertEqual(ds_path, ds_path_norm)

    def test_datatype_of_artifact(self):
        ds005 = load_dataset(DS005_DIR)
        anat_files = ds005.query(scope="raw", sub="01", suffix="T1w")
        assert len(anat_files) == 1
        assert anat_files[0].datatype is None

        ds005 = load_dataset(DS005_DIR, DatasetOptions(infer_artifact_datatype=True))
        anat_files = ds005.query(scope="raw", sub="01", suffix="T1w")
        assert len(anat_files) == 1
        assert anat_files[0].datatype == "anat"

    

if __name__ == '__main__':
    unittest.main()





import logging
import os

FILE_READERS = {}
FILE_WRITERS = {}

LOGGER = logging.getLogger(__file__)


def parse_bids_name(name: str):
    """Parses a given string (file name) according to the BIDS naming scheme.

    Parameters
    ----------
    name
        The file name to parse. If a full path (with path separators), the path segments will be ignored.

    Returns
    -------
    dict
        A dictionary describing the BIDS naming components.

    Examples
    --------

    >>> bids_obj = parse_bids_name("sub-11_task-mixedgamblestask_run-02_bold.nii.gz")
    {'entities': {'sub': '11', 'task': 'mixedgamblestask', 'run': '02'}, 'suffix': 'bold', 'extension': '.nii.gz'}

    """
    base_name = os.path.basename(name)
    parts = base_name.split(os.extsep, 1)
    if len(parts) != 2:
        # if extension missing, then not a valid BIDS file
        return None

    extension = os.extsep + parts[1]
    underscore_parts = parts[0].split('_')

    if len(underscore_parts) < 2:
        return None

    # last segment must be suffix
    suffix = underscore_parts[-1]
    if '-' in suffix:
        return None

    entities = {}
    for i in range(0, len(underscore_parts) - 1):
        dash_parts = underscore_parts[i].split('-')
        if len(dash_parts) < 2:
            # not a key-value pair
            return None
        entities[dash_parts[0]] = dash_parts[1]

    return {
        'entities': entities,
        'suffix': suffix,
        'extension': extension
    }


def load_contents(file_path, return_type: str = None):
    """Loads the contents of the provided file path.

    Parameters
    ----------
    file_path :
        the file path to load contents from
    return_type:
        A hint to consider when deciding how to load the contents of the provided file.
        For example, to load a TSV file as a pandas DataFrame the return_type should be 'dataframe',
        to load a numpy ndarray, the return_type should be 'ndarray'.
        It is up to the registered file handlers to correctly interpret the return_type.

    Returns
    -------
        The result depends on the extension of the file name.
        For example, a .json file may be returned as an ordinary Python dict or a .txt as a str value.

    """
    if not os.path.exists(file_path):
        return None
    reader = None
    file_name = os.path.basename(file_path)
    parts = os.path.splitext(file_name)
    if len(parts) > 1:
        extension = parts[-1][1:]
        if extension in FILE_READERS:
            reader = FILE_READERS[extension]
    if reader is None:
        LOGGER.debug("No reader found for file '%s', defaulting to 'txt' file reader" % file_name)
        reader = FILE_READERS['txt']
    if reader is None:
        raise ValueError('No file reader registered to load file %s' % file_path)
    return reader(file_path, return_type=return_type)


def write_contents(file_path: str, contents):
    """Writes the provided contents to the target file path using a registered file writer.

    A valid file writer may be inferred by the file's extension and/or the given contents object.
    If no file writer is found for the given file, a `ValueError` is raised.

    Parameters
    ----------
    file_path:
        The file path to write to.
    contents:
        The contents to write to the target file.

    """
    writer = None
    parts = os.path.splitext(file_path)
    if len(parts) > 1:
        extension = parts[-1][1:]
        if extension in FILE_WRITERS:
            writer = FILE_WRITERS[extension]

    if not writer:
        raise ValueError("No file writer registered for file: %s" % file_path)

    writer(file_path, contents)


def deepupdate(target, src):
    """Deep update target dict with src
    For each k,v in src: if k doesn't exist in target, it is deep copied from
    src to target. Otherwise, if v is a list, target[k] is extended with
    src[k]. If v is a set, target[k] is updated with v, If v is a dict,
    recursively deep-update it.

    Examples:
    >>> t = {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi']}
    >>> deepupdate(t, {'hobbies': ['gaming']})
    >>> print t
    {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi', 'gaming']}

    Copyright Ferry Boender, released under the MIT license.
    """
    import copy
    for k, v in src.items():
        if type(v) == list:
            if k not in target:
                target[k] = copy.deepcopy(v)
            else:
                target[k].extend(v)
        elif type(v) == dict:
            if k not in target:
                target[k] = copy.deepcopy(v)
            else:
                deepupdate(target[k], v)
        elif type(v) == set:
            if k not in target:
                target[k] = v.copy()
            else:
                target[k].update(v.copy())
        else:
            target[k] = copy.copy(v)


def fetch_dataset(dataset_id: str, output_dir='~/.ancp-bids/datasets'):
    """Downloads and extracts an ancpBIDS  test dataset from Github.

    Parameters
    ----------
    dataset_id :
        The dataset ID of the ancp-bids-datasets github repository.
        See `https://github.com/ANCPLabOldenburg/ancp-bids-dataset` for more details.

    output_dir :
        The output directory to download and extract the dataset to.
        Default is to write to user's home directory at `~/.ancp-bids/datasets`

    Returns
    -------
        The path of the extracted dataset.
    """
    output_dir = os.path.expanduser(output_dir)
    output_dir = os.path.abspath(os.path.normpath(output_dir))
    output_path = os.path.join(output_dir, dataset_id)
    if os.path.exists(output_path):
        return output_path

    os.makedirs(output_path)

    download_file = f'{dataset_id}-testdata.zip'
    download_path = os.path.join(output_dir, download_file)

    if os.path.exists(download_path):
        return output_path

    url = f'https://github.com/ANCPLabOldenburg/ancp-bids-dataset/raw/main/{download_file}'
    import urllib.request, zipfile
    with urllib.request.urlopen(url) as dl_file:
        with open(download_path, 'wb') as out_file:
            out_file.write(dl_file.read())
    z = zipfile.ZipFile(download_path)
    z.extractall(output_dir)
    return output_path


def resolve_segments(root_folder, path_, last_seg_file=False, create_if_missing=False):
    context = root_folder
    if path_:
        normalized_path = os.path.normpath(path_)
        path_segments = normalized_path.split(os.sep)
        path_segments = list(filter(lambda s: s != '', path_segments))
        if last_seg_file:
            path_ = path_segments[-1]
            path_segments = path_segments[:-1]

        for seg in path_segments:
            context = context.get_folder(seg)
            if not context:
                if create_if_missing:
                    context.create_folder(seg)
                else:
                    break
    return context, path_


def convert_to_relative(dataset, path):
    if path and path.startswith(dataset.base_dir_):
        path = os.path.normpath(path)
        path = path[len(dataset.base_dir_):]
        path_segments = path.split(os.sep)
        path_segments = list(filter(lambda s: s != '', path_segments))
        path = os.path.join(*path_segments)
    return path
